/* -*- compile-command: "msdev modeler.dsp /make \"Modeler - Win32 Debug\"" -*-
 */

#include "impressionist.h"
#include "ImpBrush.h"
#include "DotBrush.h"
#include <math.h>

extern "C" {
    int __declspec( dllexport ) Impressionist_Init( Tcl_Interp* interp );
    BOOL __declspec( dllexport ) WINAPI DllMain( HINSTANCE, DWORD, LPVOID );
}


HINSTANCE g_hInstance = NULL;
Tcl_Interp* g_interp = NULL;


CImpBrush* g_pCurrentBrush;


char command_buffer[1024];
char convert_buffer[1024];





#define BORDER 10



/* Name of our window class */
static char g_szPaintName[] = "Paint";

/* Handle to our window */
static HWND g_PaintWnd = NULL;
/* Handle to this instance of the .dll */



/* Dimensions of the window */
GLsizei g_PaintWndWidth, g_PaintWndHeight;

/* Dimensions of the painting */
GLsizei	g_PaintWidth, g_PaintHeight;

/* Location of the painting in the parent window */
GLsizei g_PaintStartX, g_PaintStartY;

/* Location of the bitmap in the parent window */
GLsizei g_OrigLocX, g_OrigLocY;
 
/* This is how we store the bitmaps */
unsigned char* g_ucBitmap = NULL;    //The original drawing
unsigned char* g_ucPainting = NULL;  //Our canvas


int g_nBrushCount = 0;
CImpBrush** g_pBrushes = NULL;
BOOL g_bSelectRedraw = FALSE;

/* DllMain
 * 
 * every DLL must provide DllMain.  ours just stores the instance handle,
 * so we can use it to create a window later.
 */

BOOL WINAPI DllMain( HINSTANCE hInst, DWORD dwReason, LPVOID reserved )
{
    g_hInstance = hInst;
    return 1;
}

/* Modeler_Init
 *
 * Tcl calls this function when the DLL is first loaded, to initialize
 * the package(s) contained in the DLL.
 */

int Impressionist_Init( Tcl_Interp* interp )
{
	WNDCLASSEX paintClass;

    /* initialize the UI control package. */
    if ( Controls_Init( interp ) != TCL_OK )
	return TCL_ERROR;


    /* register the window classes with Win32.  this is all standard,
       except we reserve 4 extra bytes in the window structure
       (cbWndExtra) to keep a copy of the Tcl interpreter pointer,
       which we'll need to eval the draw and resize objects. */
 	paintClass.cbSize = sizeof( paintClass );
    paintClass.style = CS_HREDRAW | CS_VREDRAW;
    paintClass.lpfnWndProc = PaintWndProc;
    paintClass.cbClsExtra = 0;
    paintClass.cbWndExtra = 4;
    paintClass.hInstance = g_hInstance;
    paintClass.hIcon = LoadIcon( NULL, IDI_APPLICATION );
    paintClass.hCursor = LoadCursor( NULL, IDC_ARROW );
    paintClass.hbrBackground = (HBRUSH) (COLOR_WINDOW+1);
    paintClass.lpszMenuName = NULL;
    paintClass.lpszClassName = g_szPaintName;
    paintClass.hIconSm = LoadIcon( NULL, IDI_APPLICATION );
    RegisterClassEx( &paintClass );

    /* create new commands in the Tcl interpreter for creating,
       destroy, and invalidating (forcing redraw) the opengl
       window. */
    Tcl_CreateObjCommand( interp, "::opengl::create", OpenGL_CreateCmd, NULL, NULL );
    Tcl_CreateObjCommand( interp, "::opengl::destroy", OpenGL_DestroyCmd, NULL, NULL );
    Tcl_CreateObjCommand( interp, "::opengl::invalidate", OpenGL_InvalidateCmd, NULL, NULL );
	Tcl_CreateObjCommand( interp, "::opengl::SaveBitmap", SaveBitmap, NULL, NULL );


	/* Create the brushes */
	g_nBrushCount = 1;
	g_pBrushes = new CImpBrush* [g_nBrushCount];
	g_pBrushes[0] = new CDotBrush("Points" );   
	
	g_pCurrentBrush = g_pBrushes[0];

	/* Make the controls */
	CreateControlMenu(interp);

    return TCL_OK;
}



/* This is where we setup the behavior we want our OpenGL applcation to have.  
 * To do anything, this function must be called after a call to wglMakeCurrent
 * (which specifies an OpenGL window to draw to). Luckily for you, it is.
 */
void InitializeGL()
{
	/* Tell OpenGL to clear the screen to a gray color */
	glClearColor(0.7f, 0.7f, 0.7f, 1.0);

	/* We're only using 2-D, so turn off depth */
	glDisable( GL_DEPTH_TEST );

	/* Tell openGL to read from the front buffer when capturing
	 * out paint strokes */
	glReadBuffer( GL_FRONT );

}

/* 
 * Create our impressionist controls. This is where new controls should go. 
 */
void CreateControlMenu(Tcl_Interp* interp)
{
	scale( interp, SIZE,  "Size:", 1.0, 25.0, 1.0, 15.0);
	/* TEMP FIX: Tcl refuses to acknowledge a control until you've changed
	 *  its value at least once. Hopefully your TA will figure out what's up
	 *  with that and fix it. For now, just set it immediately after creation.*/
	set_control_i(SIZE, 15);

	/* HOW TO ADD NEW BRUSHES
	 *  Simply add a name between the "Dot" and NULL parameters (this function takes
	 *  a variable number of parameters). To figure out which button is selected, 
	 *  use get_control_i, which will return an integer denoting the index of the
	 *  button. "Dot" returns 0, the next brush you add will return 1, etc...
	 */
	radiobutton(interp, BRUSHES, "Brushes", 0, "Dot", NULL);
}




/* This function handles all of the events that our window receives. */
LRESULT CALLBACK PaintWndProc( HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam )
{

	/* a DC is a Win32 "device context".  it is a handle to a Win32
       data structure needed for drawing in a window. */
    static HDC hdc;
    /* an RC is an OpenGL "rendering context".  it is a handle to an
       OpenGL data structure needed for drawing in a window. */
    static HGLRC hrc;

	POINT mouse;
	POINT canvas; // translate the mouse location to canvas coordinates 

	unsigned char *paintstart, *bitstart;  

    PAINTSTRUCT ps;

	static int i = 0;
    
    switch( iMsg )
    {

	case WM_RBUTTONDOWN:

		/* Win32 informs us that the right button has been depressed */
		hdc = GetDC(hwnd);
		::wglMakeCurrent(hdc, hrc);

		g_pCurrentBrush = g_pBrushes[get_control_i(BRUSHES)];
		/* Get the location of the mouse click in the window */
		mouse.x = (int)LOWORD( lParam );
		mouse.y = g_PaintWndHeight - (int)HIWORD( lParam );

		/* We only want to redraw the screen in MOUSEMOVE if
		 * this function actually does anything */
		g_bSelectRedraw = g_pCurrentBrush->SelectBegin(mouse);

		return 0;


	case WM_RBUTTONUP:

		/* Win32 has informed us that the right button has been released */
		hdc = GetDC(hwnd);
		::wglMakeCurrent(hdc, hrc);

		mouse.x = (int)LOWORD( lParam );
		mouse.y = g_PaintWndHeight - (int)HIWORD( lParam );

		g_pCurrentBrush->SelectEnd(mouse);

		::wglMakeCurrent( NULL, NULL );

		/* Tell the window to redraw */
		InvalidateRect(hwnd, NULL, NULL);
		return 0;


	case WM_LBUTTONDOWN:

		/* Win32 informs us that the left button has been depressed */
		if (g_ucPainting == NULL) return 0;

		hdc = GetDC(hwnd);

		/* Get the location of the mouse click, and
		 * the translated coordinates into canvas space */
		mouse.x = (int)LOWORD( lParam );
		mouse.y = g_PaintWndHeight - (int)HIWORD( lParam );
		canvas.x = mouse.x - g_PaintStartX;
		canvas.y = mouse.y - g_PaintStartY;
		
		wglMakeCurrent( hdc, hrc );
		glDrawBuffer(GL_FRONT);


		/* Get the current brush from our radio box */
		g_pCurrentBrush = g_pBrushes[get_control_i(BRUSHES)];

		g_pCurrentBrush->BrushBegin( mouse, canvas);		

		return 0;


    case WM_LBUTTONUP:

		/* Win32 has informed us that the left button has been released*/
		hdc = GetDC(hwnd);
		::wglMakeCurrent(hdc, hrc);



		mouse.x = (int)LOWORD( lParam );
		mouse.y = g_PaintWndHeight - (int)HIWORD( lParam );

		canvas.x = mouse.x - g_PaintStartX;
		canvas.y = mouse.y - g_PaintStartY;

		g_pCurrentBrush->BrushEnd(mouse, canvas);

		Debug("Read from %d, %d a length of (%d, %d)\n", g_PaintStartX, g_PaintStartY, 
			g_PaintWidth, g_PaintHeight);
		glReadBuffer( GL_FRONT );
		glPixelStorei( GL_PACK_ALIGNMENT, 1 );
		glPixelStorei( GL_PACK_ROW_LENGTH, g_PaintWidth );
		glReadPixels( g_PaintStartX, g_PaintStartY, g_PaintWidth, g_PaintHeight, 
			          GL_RGB, GL_UNSIGNED_BYTE, g_ucPainting);

		::wglMakeCurrent( NULL, NULL );
		
		/* Tell the window to redraw */

		InvalidateRect(hwnd, NULL, NULL);
		return 0;


	case WM_MOUSEMOVE: 

		/*Win32 informs us that the mouse has moved */

		mouse.x = (int)LOWORD( lParam );
		//Invert the y to go from WIN32 y to OpenGL y 
		mouse.y = g_PaintWndHeight - (int)HIWORD( lParam );
		canvas.x = mouse.x - g_PaintStartX;
		canvas.y = mouse.y - g_PaintStartY;


		hdc = GetDC(hwnd);
		::wglMakeCurrent(hdc, hrc);

		if (wParam == MK_LBUTTON)
		{
		   g_pCurrentBrush->BrushMove(mouse, canvas);
		}
		else if (wParam == MK_RBUTTON)
		{
			if ( g_bSelectRedraw )
			{	
				glDrawBuffer( GL_BACK );
				glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
				glPixelStorei( GL_UNPACK_ROW_LENGTH, g_PaintWidth );
				glRasterPos2i(g_PaintStartX, g_PaintStartY);
				glDrawPixels( g_PaintWidth, g_PaintHeight, GL_RGB, 
							  GL_UNSIGNED_BYTE, g_ucPainting);
			}
			g_pCurrentBrush->SelectMove(mouse);
			if ( g_bSelectRedraw )
			{
				::SwapBuffers( hdc );
			}
		}

		wglMakeCurrent(NULL, NULL);

		return 0;

	case WM_CREATE:
	    /* when the window is created, create an OpenGL RC to go
	       along with it.  we will keep this RC around for the
	       entire lifetime of the window. */
	    hrc = SetupOpenGL( hwnd );
	    return 0;

	case WM_ERASEBKGND:
	    /* we will erase the window ourselves when it is redrawn.
	       tell Win32 NOT to erase the window for us, by
	       returning true. */
	    return 1;

	case WM_PAINT:
	    /* Win32 has requests that the window be redrawn. */

	    /* calling BeginPaint is obligatory for handling a
	       WM_PAINT message.  it returns a DC to let us draw in
	       the window. */
	    hdc = BeginPaint( hwnd, &ps );

		/* Tell OpenGL which window to draw to. Without
		 * this, none of your OpenGL calls will do anything,
		 * and you won't know why. */
	    wglMakeCurrent( hdc, hrc );

		/* Draw everything to the back buffer */
		glDrawBuffer(GL_BACK);

		/* Clear out the back buffer */
		glClear( GL_COLOR_BUFFER_BIT );

		if(g_PaintStartY < 0)
		{
		 paintstart = g_ucPainting + (3*g_PaintWidth * abs(g_PaintStartY));
		 bitstart = g_ucBitmap + (3*g_PaintWidth * abs(g_OrigLocY));
		}
		else 
		{
			paintstart = g_ucPainting;
			bitstart = g_ucBitmap;
		}
		

		/* Draw the canvas */
		glRasterPos2i(g_PaintStartX, max(0, g_PaintStartY));
		glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
		glPixelStorei(GL_UNPACK_ROW_LENGTH, g_PaintWidth);
		glDrawPixels(g_PaintWidth, g_PaintHeight, GL_RGB, GL_UNSIGNED_BYTE, paintstart);

		/* Draw the original bitmap to the right of the canvas */
		glRasterPos2i(g_OrigLocX, max(0, g_OrigLocY));
		glDrawPixels(g_PaintWidth, g_PaintHeight, GL_RGB, GL_UNSIGNED_BYTE, bitstart);
	
		/* Tell OpenGL to draw everything */
		glFlush();
		SwapBuffers(hdc);
		
		/* Set the current drawing window to NULL. 
		 * This is done to prevent accidental drawing
		 * to our window */
		wglMakeCurrent(NULL, NULL);
	  
		
	    /* each BeginPaint call must be matched with an EndPaint
	       call to clean up. */
	    EndPaint( hwnd, &ps );

	    return 0;


	case WM_SIZE:
	    /* Win32 informs us that the window has been resized. */

		
	    /* the window's new size is passed to us in lParam. */
	    g_PaintWndWidth = (GLsizei)LOWORD( lParam );
	    g_PaintWndHeight = (GLsizei)HIWORD( lParam );

		/* Re-center the images and store their new starting locations */
		g_PaintStartX = max(0, (g_PaintWndWidth/4) - g_PaintWidth/2);
		g_PaintStartY = (g_PaintWndHeight/2) - (g_PaintHeight/2);

		g_OrigLocX =  max(g_PaintStartX + g_PaintWidth, (3*g_PaintWndWidth)/4 - g_PaintWidth/2);
		g_OrigLocY = g_PaintStartY;

	//	Debug("Window Width: %d\n Window Height: %d\nPaint Width: %d\nPaint Height: %d, StartX: %d,StartY: %d",
	//		   g_PaintWndWidth, g_PaintWndHeight, g_PaintWidth, g_PaintHeight, 
	//		   g_PaintStartX, g_PaintStartY);
	
	    /* get a DC for the window, and tell OpenGL to use it with the saved RC. */
	    hdc = GetDC( hwnd );
	    wglMakeCurrent( hdc, hrc );

		glViewport(0, 0, g_PaintWndWidth, g_PaintWndHeight);
	   	glMatrixMode( GL_PROJECTION );
		glLoadIdentity();
		gluOrtho2D( 0, g_PaintWndWidth, 0, g_PaintWndHeight );

		glMatrixMode( GL_MODELVIEW );
		glLoadIdentity();
		glTranslatef( 0.375, 0.375, 0.0 );

	    /* tell OpenGL to stop drawing in the window and release the DC. */
	    wglMakeCurrent( NULL, NULL );
	    ReleaseDC( hwnd, hdc );

	    /* all done! */
		InvalidateRect(hwnd, NULL, NULL);
	    return 0;

	case WM_CLOSE:
	    /* Win32 requests that the window be closed. */

	    /* delete the window. */
	    DestroyWindow( hwnd );
	    g_PaintWnd = NULL;
	    
	    return FALSE;
    }

    /* any other messages that we don't handle will be handled by
       DefWindowProc, which does the expected defaulty sort of
       things. */
    return DefWindowProc( hwnd, iMsg, wParam, lParam );
   
}


/* 
 * Receives a point in canvas coordinates, and returns
 * an index into the bitmap at that location. 
 */
GLubyte* GetOriginalPixel( int x, int y )
{
	/* Make sure the point is somewhere on our picture */
	if (x >= g_PaintWidth) x = g_PaintWidth - 3;
	if (x < 0) x = 0;
	if (y >= g_PaintHeight) y = g_PaintHeight - 3;
	if (y < 0) y = 0;
	
	/* Return an index into the array containing our bitmap */
	return (GLubyte*)(g_ucBitmap + 3 * (y*g_PaintWidth + x));
}

GLubyte* GetOriginalPixel( const POINT p )
{
	return GetOriginalPixel( p.x, p.y );
}


/* Debug
 *
 * Debug works just like printf(), except that it prints output in a 
 * message box.
 */

void Debug( char* fmt, ... )
{
    char buffer[1024];
    va_list list;

    va_start( list, fmt );
    vsprintf( buffer, fmt, list );
    va_end( list );

    OutputDebugString(buffer);
}

/* Load a bitmap from disk */
bool LoadBitmap()
{
		FILE* file;
		DWORD pos;
		
		BITMAPFILEHEADER bmfh;
		BITMAPINFOHEADER bmih;
		int i;
		char fname[100];
		fname[0] = '\0';

		OPENFILENAME open;
		
		open.lStructSize = sizeof(OPENFILENAME);
		open.hwndOwner = NULL;
		open.lpstrFilter = "Bitmap (*.bmp)\0*.bmp\0\0";
		open.nFilterIndex = 0;
		open.lpstrCustomFilter = NULL;
		open.nFilterIndex = NULL;
		open.lpstrFile = fname;
		open.nMaxFile = 2048000;
		open.lpstrFileTitle = NULL;
		open.lpstrInitialDir = NULL;
		open.lpstrTitle = "Bmp loader";
		open.Flags = NULL;
		open.nFileOffset = NULL;
		open.nFileExtension = NULL;
		open.lpstrDefExt = "bmp";
		open.lCustData = NULL;
		open.lpfnHook = NULL;
		open.lpTemplateName = NULL;

		GetOpenFileName(&open);

		if (open.lpstrFile[0] == '\0') return false;

	    file = fopen(open.lpstrFile, "r");
		
		fread( &bmfh, sizeof(BITMAPFILEHEADER), 1, file );
		pos = bmfh.bfOffBits;
		fread( &bmih, sizeof(BITMAPINFOHEADER), 1, file );

		if ( bmih.biBitCount != 24 ) {
			MessageBox( NULL, "Only 24-bit BMPs are supported.", NULL, MB_OK | MB_ICONEXCLAMATION );
			return false;
		}
		if ( bmih.biCompression != BI_RGB ) {
			MessageBox( NULL, "Compressed BMPs are not supported.", NULL, MB_OK | MB_ICONEXCLAMATION );
			return false;
		}

		/* Return to the start of the file */
		fseek(file,  pos, SEEK_SET);

		int bytes, pad;
		bytes = bmih.biWidth * 3;
		pad = 0;
		if ( bytes % 4 != 0 )
		{
			pad = 4 - (bytes % 4);
			bytes += pad;
		}
		bytes *= bmih.biHeight;


		if ( g_ucBitmap ) { delete [] g_ucBitmap; g_ucBitmap = NULL; }
		if ( g_ucPainting ) { delete [] g_ucPainting; g_ucPainting = NULL; }
		

		g_ucBitmap = new unsigned char[bytes];
		g_ucPainting = new unsigned char[bytes];
		
		/* Make the canvas white */
		memset(g_ucPainting, 255, bytes);


		/* Read in the actual bitmap */
		fread( g_ucBitmap, bytes, 1, file);		
					
		fclose(file);

		/* Store the dimensions of the painting */
		g_PaintWidth  =  bmih.biWidth;
		g_PaintHeight  = bmih.biHeight;
		
		/* Make up the dimensions of the window */
		g_PaintWndWidth =  (int)(g_PaintWidth * 2.1) + BORDER;
		g_PaintWndHeight =  g_PaintHeight + 5*BORDER;

		int j;
		j = 0;
		unsigned char temp;
		unsigned char* in;
		unsigned char* out;

		// remove padding and twiddle byte order of incoming bitmap. OpenGL expects
		// our pixels to be in a certain R-G-B order, and if they're not we're doomed!

		in = g_ucBitmap;
		out = g_ucBitmap;

		for ( j = 0; j < g_PaintHeight; ++j )
		{
			for ( i = 0; i < g_PaintWidth; ++i )
			{
				out[1] = in[1];
				temp = in[2];
				out[2] = in[0];
				out[0] = temp;

				in += 3;
				out += 3;
			}
			in += pad;
		}

	return true;		
}

/* Saves a bitmap to disk */
int SaveBitmap(ClientData clientData, Tcl_Interp* interp,
			  int objc, Tcl_Obj* const objv[])
{
		FILE* file;
		OPENFILENAME open;
		char fname[100];
		
		BITMAPFILEHEADER bmfh;
		BITMAPINFOHEADER bmih;

		fname[0] = '\0';

		open.lStructSize = sizeof(OPENFILENAME);
		open.hwndOwner = NULL;
		open.lpstrFilter = "Bitmap (*.bmp)\0*.bmp\0\0";
		open.nFilterIndex = 0;
		open.lpstrCustomFilter = NULL;
		open.nFilterIndex = NULL;
		open.lpstrFile = fname;
		open.nMaxFile = 2048000;
		open.lpstrFileTitle = NULL;
		open.lpstrInitialDir = NULL;
		open.lpstrTitle = "Bmp loader";
		open.Flags = NULL;
		open.nFileOffset = NULL;
		open.nFileExtension = NULL;
		open.lpstrDefExt = "bmp";
		open.lCustData = NULL;
		open.lpfnHook = NULL;
		open.lpTemplateName = NULL;


		GetSaveFileName(&open);
	

		int bytes, pad;
		bytes = g_PaintWidth * 3;
		pad = (bytes%4) ? 4-(bytes%4) : 0;
		bytes += pad;
		bytes *= g_PaintHeight;

		bmfh.bfType = 0x4d42;    // "BM"
		bmfh.bfSize = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + bytes;
		bmfh.bfReserved1 = 0;
		bmfh.bfReserved2 = 0;
		bmfh.bfOffBits = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);

		bmih.biSize = sizeof(BITMAPINFOHEADER);
		bmih.biWidth = g_PaintWidth;
		bmih.biHeight = g_PaintHeight;
		bmih.biPlanes = 1;
		bmih.biBitCount = 24;
		bmih.biCompression = BI_RGB;
		bmih.biSizeImage = 0;
		bmih.biXPelsPerMeter = (int)(100 / 2.54 * 72);
		bmih.biYPelsPerMeter = (int)(100 / 2.54 * 72);
		bmih.biClrUsed = 0;
		bmih.biClrImportant = 0;

		file = fopen(open.lpstrFile, "w");
		
		fwrite( &bmfh, sizeof(BITMAPFILEHEADER), 1, file );
		fwrite( &bmih, sizeof(BITMAPINFOHEADER), 1, file );

		bytes /= g_PaintHeight;
		unsigned char* scanline = new unsigned char [bytes];
		for ( int j = 0; j < g_PaintHeight; ++j )
		{
			memcpy( scanline, g_ucPainting + j*3*g_PaintWidth, bytes );
			for ( int i = 0; i < g_PaintWidth; ++i )
			{
				unsigned char temp = scanline[i*3];
				scanline[i*3] = scanline[i*3+2];
				scanline[i*3+2] = temp;
			}
			fwrite( scanline, bytes, 1, file );
		}
		delete [] scanline;
		fclose(file);
	return TCL_OK;
}


/* OpenGL_InvalidateCmd
 *
 * when this function is called from Tcl, invalidate the output window
 * so that Win32 will ask to redraw it.
 */
int OpenGL_InvalidateCmd( ClientData clientData, Tcl_Interp* interp,
			  int objc, Tcl_Obj* const objv[] )
{
    if ( objc != 1 )
    {
	Tcl_SetResult( interp, "wrong # args: should be \"destroy\"",
		       TCL_STATIC );
	return TCL_ERROR;
    }

	if (g_PaintWnd)
		InvalidateRect( g_PaintWnd, NULL, FALSE );

    
    return TCL_OK;
}

/* OpenGL_DestroyCmd
 *
 * when this function is called from Tcl, close the output window.
 */


    
int OpenGL_DestroyCmd( ClientData clientData, Tcl_Interp* interp,
		       int objc, Tcl_Obj* const objv[] )
{
    if ( objc != 1 )
    {
	Tcl_SetResult( interp, "wrong # args: should be \"destroy\"",
		       TCL_STATIC );
	return TCL_ERROR;
    }

	/* Tell the window to erase itself */
	if ( g_PaintWnd )
	SendMessage( g_PaintWnd, WM_CLOSE, 0, 0 );

    return TCL_OK;
}

/* OpenGL_CreateCmd
 *
 * when this function is called from Tcl, create the output window.
 * 
 */

int OpenGL_CreateCmd( ClientData clientData, Tcl_Interp* interp,
		      int objc, Tcl_Obj* const objv[] )
{


	/* Load the bitmap. If it fails, exit (but don't alarm Tcl 
	 * by sending back an error) */
	if (!LoadBitmap()) return TCL_OK;
	
    /* create a Win32 window.  this is all standard, except that
       OpenGL requires the WS_CLIPCHILDREN and WS_CLIPSIBLINGS
	   flags. */
	g_PaintWnd = CreateWindow( g_szPaintName,
			   "Paint window", 
			   WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN |
			   WS_CLIPSIBLINGS, 
			   CW_USEDEFAULT, CW_USEDEFAULT,
			   g_PaintWndWidth, g_PaintWndHeight,
			   NULL, NULL, g_hInstance, NULL );
	
	
	
    /* store a copy of the interpreter pointer in the window
       structure. */
    SetWindowLong( g_PaintWnd, 0, (LONG)interp );

    /* show the window on the screen. */
    ShowWindow( g_PaintWnd, SW_SHOWNA );

	/* Tell the window to re-draw itself */
    UpdateWindow( g_PaintWnd );


    return TCL_OK;
}

void InitializeGL();

/* SetupOpenGL
 *
 * when the output window is created, we must create an OpenGL
 * rendering context (RC) for it.
 */

HGLRC SetupOpenGL( HWND hwnd )
{
    /* the pixel format descriptor tells OpenGL what features we need
       to be supported in the window. */
    static PIXELFORMATDESCRIPTOR pfd = {
	sizeof( PIXELFORMATDESCRIPTOR ),
	1,
	/* request a double-buffered window */
	PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
	PFD_TYPE_RGBA,
	/* request a 24-bit color window */
	24,
	0, 0, 0,
	0, 0, 0,
	0, 0, 
	0, 0, 0, 0, 0,
	/* request a 16-bit depth buffer */
	16,
	0, 
	0,
	PFD_MAIN_PLANE,
	0,
	0,
	0,
	0
    };
    int myPixelFormatID;
    HDC hdc;
    HGLRC hrc;

    /* get a DC for the window. */
    hdc = GetDC( hwnd );

    /* look for a pixel format that matches the requests we put in
       the pixel format descriptor. */
    myPixelFormatID = ChoosePixelFormat( hdc, &pfd );
    if ( myPixelFormatID == 0 )
    {
	/* if no appropriate pixel format could be found, die rather
	   ungracefully. */
	Debug( "this system doesn't support the required visual!" );
	exit( 1 );
    }

    /* set up the window with this pixel format. */
    SetPixelFormat( hdc, myPixelFormatID, &pfd );

    /* create a rendering context for this window. */
    hrc = wglCreateContext( hdc );

	::wglMakeCurrent( hdc, hrc );

	/* Setup default OpenGL behavior */
	InitializeGL();

	::wglMakeCurrent( NULL, NULL );

    /* release the DC. */
    ReleaseDC( hwnd, hdc );

    return hrc;
}


