/* -*- compile-command: "msdev modeler.dsp /make \"Modeler - Win32 Debug\"" -*-
 */

#ifndef _MODELER_H
#define _MODELER_H

#include <windows.h>
#include <tk.h>
#include <assert.h>

#include <gl\gl.h>
#include <gl\glu.h>





/* ID codes for our controls */
#define ALPHA        1
#define SIZE         2
#define ANGLE        3
#define WIDTH        4
#define ANGLE_SELECT 5
#define BRUSHES      6



typedef struct
{
    int type;
    Tcl_Obj* widget;
} control;

/* grrr.  why does Microsoft not put the value of pi in any of its
   standard include files? */
#define M_PI 3.14159265358979323846

/* we'll use static shared buffers for constructing Tcl commands
   (which are evaluated and immediately thrown away). */
#define BUFFER_SIZE  1024
extern char command_buffer[BUFFER_SIZE];
extern char convert_buffer[BUFFER_SIZE];

/* CONVERT (used in controls.c) takes the string variable "s" and
   escapes spaces and funny characters in it according to Tcl rules,
   placing the result in convert_buffer. */
#define CONVERT(s) {int flags=0; \
                   assert( Tcl_ScanElement( s, &flags ) < BUFFER_SIZE );\
                   Tcl_ConvertElement( s, convert_buffer, flags );}
		   


/*
 * function prototypes:  anything marked with "__declspec(dllexport)"
 * is something you can call from your modeler function(s).  anything
 * without this declaration is only needed internally, and you can't
 * call it (so don't try!).
*/

/** OpenGL module **/

extern "C" {
    
    int OpenGL_CreateCmd( ClientData clientData, Tcl_Interp* interp,
			  int objc, Tcl_Obj* const objv[] );

    int OpenGL_DestroyCmd( ClientData clientData, Tcl_Interp* interp,
			   int objc, Tcl_Obj* const objv[] );

    int OpenGL_InvalidateCmd( ClientData clientData, Tcl_Interp* interp,
			      int objc, Tcl_Obj* const objv[] );

	int SaveBitmap(ClientData clientData, Tcl_Interp* interp,
					int objc, Tcl_Obj* const objv[] );
}


LRESULT CALLBACK PaintWndProc( HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam );

HGLRC SetupOpenGL( HWND hwnd );
void InitializeGL();

bool LoadBitmap(void);

void CreateControlMenu(Tcl_Interp* interp);

GLubyte *GetOriginalPixel(const POINT p);
unsigned char GetOriginalGradient( int x, int y );
unsigned char GetOriginalGradient( const POINT p );

/* Debug is like printf, but prints to a floating dialog box.*/
void __declspec( dllexport ) Debug( char* fmt, ... );



/* internal stuff, can't call these. */
int Controls_Init( Tcl_Interp* interp );
control* new_control( int, int );
void reset_controls ( void );


#define CONTROL_SCALE        1
#define CONTROL_CHECKBOX     2
#define CONTROL_RADIOBUTTON  3
#define CONTROL_MENU         4
#define CONTROL_TIMECURVE    5

/* for all controls, interp is the Tcl_Interp parameter passed to
   Model_SetupCmd, id is an integer unique to that control, and name
   is a string used for the label. */

/* slider from "min" to "max" in steps of "res", initial value "start". */
int __declspec( dllexport )
    scale( Tcl_Interp* interp, int id, char* name, double min, double max,
	   double res, double start );

/* checkbox, initial state "state". */
int __declspec( dllexport )
    checkbox( Tcl_Interp* interp, int id, char* name, int state );

/* set of radiobuttons with labels "..." (NULL-terminated).  "start" is
   the number of the one initially checked. */
int __declspec( dllexport )
    radiobutton( Tcl_Interp* interp, int id, char* name, int start, ... );

/* menu with radiobutton-type options "..." (NULL-terminated).
   "start" is the one initially checked. */
int __declspec( dllexport )
    menu( Tcl_Interp* interp, int id, char* name, int start, ... );
    

/* read the value of a control as a double, int, or boolean. */    
double __declspec( dllexport ) get_control_d( int id );
void __declspec( dllexport ) set_control_d( int id, double value );

int __declspec( dllexport ) get_control_i( int id );
void __declspec( dllexport ) set_control_i( int id, int value );

int __declspec( dllexport ) get_control_b( int id );
void __declspec( dllexport ) set_control_b( int id, int value );

#endif 


    
